/****************************************************************************
    
    unicode.hpp - Utility functions for encoding and decoding UTF-8 strings
    
    Copyright (C) 2006-2007 Lars Luthman <lars.luthman@gmail.com>
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 3 of the License, or
    (at your option) any later version.
    
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 01222-1307  USA

****************************************************************************/

#ifndef UNICODE_HPP
#define UNICODE_HPP

#include <iostream>
#include <string>
#include <utility>


using namespace std;
using namespace boost::spirit::classic;


template <typename ScannerT>
pair<unsigned int, unsigned int> decode_utf8(ScannerT const& scan) {
  
  typedef typename ScannerT::iterator_t iterator_t;
  typedef typename ScannerT::value_t value_t;
  
  if (scan.at_end())
    return make_pair(0, 0);
  
  iterator_t iter(scan.first);
  
  value_t ch = *scan.first;
  
  unsigned int size;
  unsigned int codepoint = 0;
  
  // one byte char
  if ((unsigned char)ch < 0x80) {
    size = 1;
    codepoint = ch;
  }
  
  // two byte char
  else if ((unsigned char)ch >= 192 && (unsigned char)ch < 224) {
    codepoint = ((unsigned char)ch && 0x1F) << 6;
    ++iter;
    if (iter == scan.last)
      return make_pair(0, 0);
    codepoint += (unsigned char)(*iter) && 0x3F;
    size = 2;
  }
  
  // three byte char
  else if ((unsigned char)ch >= 224 && (unsigned char)ch < 240) {
    codepoint = ((unsigned char)ch && 0x0F) << 12;
    ++iter;
    if (iter == scan.last)
      return make_pair(0, 0);
    codepoint += ((unsigned char)(*iter) && 0x3F) << 6;
    ++iter;
    if (iter == scan.last)
      return make_pair(0, 0);
    codepoint += ((unsigned char)(*iter) && 0x3F);
    size = 3;
  }
  
  // four byte char
  else if ((unsigned char)ch >= 240 && (unsigned char)ch < 248) {
    codepoint = ((unsigned char)ch && 0x07) << 18;
    ++iter;
    if (iter == scan.last)
      return make_pair(0, 0);
    codepoint += ((unsigned char)(*iter) && 0x3F) << 12;
    ++iter;
    if (iter == scan.last)
      return make_pair(0, 0);
    codepoint += ((unsigned char)(*iter) && 0x3F) << 6;
    ++iter;
    if (iter == scan.last)
      return make_pair(0, 0);
    codepoint += ((unsigned char)(*iter) && 0x3F);
    size = 4;
  }
  
  // this is not allowed in Turtle!
  else {
    assert(!"Unexpected parse error");
  }
  
  return make_pair(size, codepoint);
}

                                                     
class uch_p : public parser<uch_p> {
public:
  
  uch_p(unsigned int codepoint = 0) 
    : valid(codepoint) {
    
  }
  
  unsigned int valid;
  
  typedef uch_p self_t;
  
  template <typename ScannerT> struct result {
    typedef typename match_result<ScannerT, unsigned int>::type type;
  };
  
  template <typename ScannerT> typename parser_result<self_t, ScannerT>::type
  parse(ScannerT const& scan) const {
    
    pair<unsigned int, unsigned int> p = decode_utf8(scan);
    
    if (p.first > 0 && p.second == valid) {
      typedef typename ScannerT::iterator_t iterator_t;
      iterator_t save = scan.first;
      for (unsigned int i = 0; i < p.first; ++i)
        ++scan.first;
      return scan.create_match(p.first, p.second, save, scan.first);
    }
    
    return scan.no_match();
  }
  
};


class urange_p : public parser<urange_p> {
public:
  
  urange_p(unsigned int from, unsigned int to) 
    : m_from(from),
      m_to(to) {
    
  }
  
  unsigned int m_from, m_to;
  
  typedef urange_p self_t;
  
  template <typename ScannerT> struct result {
    typedef typename match_result<ScannerT, unsigned int>::type type;
  };

  template <typename ScannerT> typename parser_result<self_t, ScannerT>::type
  parse(ScannerT const& scan) const {
    
    pair<unsigned int, unsigned int> p = decode_utf8(scan);
    
    if (p.first > 0 && p.second >= m_from && p.second <= m_to) {
      typedef typename ScannerT::iterator_t iterator_t;
      iterator_t save = scan.first;
      for (unsigned int i = 0; i < p.first; ++i)
        ++scan.first;
      return scan.create_match(p.first, p.second, save, scan.first);
    }
    
    return scan.no_match();
  }
  
};


#endif
